"""Class for performing plots (e.g. plotting mean, Gaussian fittings)"""

"""___Built-In Modules___"""
from nectar_msi.gaussian_fitting.gaussian_fitting_definitions import GaussianFittingDefinitions

"""___Third-Party Modules___"""
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from mpl_toolkits.axes_grid1 import make_axes_locatable
import os

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class Plotting:
    def __init__(self):
        self.gaussian_def = GaussianFittingDefinitions()
        pass

    def plot_mean_spectrum(self, data_spectrum):
        """
        Function to plot the mean spectrum

        :param data_spectrum: data spectrum to plot
        :type data_spectrum: DataSpectrum
        :return: None
        :rtype: None
        """

        plt.plot(data_spectrum.mzs, data_spectrum.intensities)
        plt.title('Mean spectrum')
        plt.xlabel('m/z')
        plt.ylabel('Intensity')
        plt.show()
        plt.clf()

    def plot_noise_spectrum(self, mean_spectrum, intensities_peaks, intensities_noise):
        """
        Function to show the separation between signal and noise in the mean spectrum

        :param mean_spectrum: mean spectrum in which the noise correction has been applied
        :type mean_spectrum: DataSpectrum
        :param intensities_peaks: intensities masked as signal
        :type intensities_peaks: array of floats
        :param intensities_noise: intensities masked as noise
        :type intensities_noise: array of floats
        :return: None
        :rtype: None
        """

        plt.plot(mean_spectrum.mzs, mean_spectrum.intensities, color="black", linewidth=0.5, label='Signal')
        plt.plot(mean_spectrum.mzs, intensities_peaks, color="black", linewidth=0.5)
        plt.plot(mean_spectrum.mzs, intensities_noise, color="orange", label='Noise')
        plt.title('Noise/signal - mean spectrum')
        plt.xlabel('m/z')
        plt.ylabel('Intensity')
        plt.legend()
        plt.show()
        plt.clf()

    def plot_spectra_with_chemicalnoise(self, spectrum, intensities_new, chemical_noise):
        """
        Function to show the original mean spectrum, the subtracted modeled chemical noise and
        the corrected mean spectrum.

        :param spectrum: original mean spectrum
        :type spectrum: DataSpectrum
        :param intensities_new: intensities of the spectrum after correction
        :type intensities_new: array
        :param chemical_noise: modelled chemical noise that has been subtracted from the original spectrum
        :type chemical_noise: array
        :return: None
        :rtype: None
        """

        plt.plot(spectrum.mzs, spectrum.intensities, color="blue", label="Original", alpha=0.7)
        plt.plot(spectrum.mzs, chemical_noise, color="red", label="Chemical noise", alpha=0.6)
        plt.plot(spectrum.mzs, intensities_new, color="green", label="Corrected", alpha=0.7)
        plt.title('Chemical noise correction')
        plt.xlabel('m/z')
        plt.ylabel('Intensity')
        plt.legend()
        plt.show()
        plt.clf()

    def plot_peak_picking(self, mean_spectrum, mz_mean, peak_intensity_mean):
        """
        Function to plot the selected peaks after first noise correction

        :param mean_spectrum: spectrum used to determine the peaks
        :type mean_spectrum: DataSpectrum
        :param mz_mean: centroid values of the peaks
        :type mz_mean: array
        :param peak_intensity_mean: peaks intensity
        :type peak_intensity_mean: array
        :return: None
        :rtype: None
        """

        plt.plot(mean_spectrum.mzs, mean_spectrum.intensities, color="black")
        plt.scatter(mz_mean, peak_intensity_mean, color="red")
        plt.title('Selected peaks')
        plt.xlabel('m/z')
        plt.ylabel('Intensity')
        plt.show()
        plt.clf()

    def plot_correlation(self, mz, width, func, xaxis, path_outputs):
        """
        Function to plot the resolving power

        :param mz: peaks centroid value
        :type mz: array
        :param width: width obtained in the gaussian fitting step
        :type width: array
        :param func: linear correlation with the ordinate forced to zero
        :type func: poly1d
        :param xaxis: x-axis range of the spectrum
        :type xaxis: array
        :param path_outputs: path where to save the plot
        :type path_outputs: str
        :return: None
        :rtype: None
        """

        plt.scatter(mz, width, color="blue", alpha=0.3, s=50)
        plt.plot(xaxis, func(xaxis), linestyle="dashed", color="aqua", label="200 peaks correlation")
        plt.title("Resolving power - most intense peaks")
        plt.xlabel("Centroid")
        plt.ylabel("Width")
        plt.savefig(path_outputs + "Resolving_power.png")
        plt.clf()

    def plot_correlation_all_peaks(self, mz, width, func, func_all_peaks, xaxis, path_outputs):
        """
        Function to plot the correlation obtained with all peaks compared with the resolving power

        :param mz: peaks centroid value
        :type mz: array
        :param width: width obtained in the gaussian fitting step
        :type width: array
        :param func: linear correlation with the ordinate forced to zero
        :type func: poly1d
        :param func_all_peaks: linear correlation when using all peaks
        :type func_all_peaks: poly1d
        :param xaxis: x axis range of the spectrum
        :type xaxis: array
        :param path_outputs: path where to save the plot
        :type path_outputs: str
        :return: None
        :rtype: None
        """

        plt.scatter(mz, width, color="blue", alpha=0.3, s=50)
        plt.plot(xaxis, func(xaxis), linestyle="dashed", color="aqua", label="Resolving power")
        plt.plot(xaxis, func_all_peaks(xaxis), linestyle="dashed", color="lawngreen", label="All peaks correlation")
        plt.title("All peaks correlation")
        plt.xlabel("Centroid")
        plt.ylabel("Width")
        plt.legend(ncol=2)
        plt.savefig(path_outputs + "All_peaks_correlation.png")
        plt.clf()

    def plot_tic(self, data_cube, data_spectrum):
        """
        Function to plot the Total Ion Count image from the full data cube

        :param data_cube: data cube
        :type data_cube: DataCube
        :param data_spectrum: mean spectrum to define first and last m/z values
        :type data_spectrum: DataSpectrum
        :return: total ion counts image
        :rtype: image
        """

        image = data_cube.get_intensities_between_mzs(data_spectrum.mzs[0], data_spectrum.mzs[-1])
        im = plt.imshow(image, cmap=cm.YlGnBu_r)
        plt.colorbar(im)
        plt.show()
        plt.clf()
        return im

    def plot_sii_final_list(self, data_cube, data_mean, final_table, path_outputs, save_fig=True):
        """
        Function to create single ion images from a list of interest

        :param data_cube: data cube from where the singles ion images are created
        :type data_cube: DataCube
        :param data_mean: mean spectrum to select and plot the specific peak
        :type data_mean: DataSpectrum
        :param final_table: table with peaks to plot
        :type final_table: pandas dataframe
        :param path_outputs: path where to save the images
        :type path_outputs: str
        :param save_fig: option to save the images
        :type save_fig: bool
        :return: None
        :rtype: None
        """

        '''Creates directory to save images'''
        if save_fig:
            dirName = path_outputs + "Single_ion_images"
            if not os.path.exists(dirName):
                os.mkdir(dirName)

        '''Read table'''
        print('Creating ion images...')
        for i in range(len(final_table["meas mz"])):
            mzlow = final_table.iloc[i, 0]
            mzhigh = final_table.iloc[i, 2]
            image = data_cube.get_intensities_between_mzs(mzlow, mzhigh)
            '''define figure'''
            fig, (ax0, ax1) = plt.subplots(ncols=2)
            '''Creates single ion image'''
            im = ax0.imshow(image, cmap=cm.YlGnBu_r)
            divider = make_axes_locatable(ax0)
            cax = divider.append_axes("right", size="5%", pad=0.1)
            plt.colorbar(im, cax=cax)

            '''Creates peak image'''
            intensity_window = data_mean.intensities[np.where((data_mean.mzs > mzlow) & (data_mean.mzs < mzhigh))]
            peak = np.max(intensity_window) + 2
            ax1.plot(data_mean.mzs, data_mean.intensities, color="black")
            ax1.set_xlim((mzlow - 0.05, mzhigh + 0.05))
            ax1.set_ylim((-0.1, peak + peak*0.02))
            ax1.axvline(mzlow, color="green")
            ax1.axvline(mzhigh, color="green")
            ax1.axvline(final_table.iloc[i, 1], color="red")
            ax1.scatter(final_table.iloc[i, 1], np.max(intensity_window), color="green", marker="o")

            plt.subplots_adjust(left=0.1, bottom=None, right=None, top=None, wspace=0.3, hspace=None)
            fig.set_size_inches(14, 8)
            plt.title(str(round(final_table.iloc[i, 1], 3)) + " m/z", style='italic')

            '''Saves or shows figure'''
            if save_fig:
                plt.savefig(path_outputs + "\\Single_ion_images\\" + str(round(final_table.iloc[i, 1], 3)) + "mz.png",
                            dpi=100)
            else:
                plt.show()

            plt.close(fig)
        print('All images saved!')

    def plot_backgroundNoise(self, image, data_spectrum, mz, mzlow, mzhigh, noise_sii_background, a, mask, sn_tissue,
                             sn_background, path_outputs):
        """
        Function to plot the information related to the background noise of the peak

        :param image: image to study
        :type image: image
        :param data_spectrum: mean spectrum to read the peak
        :type data_spectrum: DataSpectrum
        :param mz: peak centroid
        :type mz: float
        :param mzlow: peak's left minima
        :type mzlow: float
        :param mzhigh:peak's right minima
        :type mzhigh: float
        :param noise_sii_background: noise in background
        :type noise_sii_background: array
        :param a: estimated noise value
        :type a: float
        :param mask: mask to identify tissue only
        :type mask: array
        :param sn_tissue: s/n tissue
        :type sn_tissue: float
        :param sn_background: s/n background
        :type sn_background: float
        :param path_outputs: path where to save the outputs
        :type path_outputs: str
        :return: None
        :rtype: None
        """

        '''Creates directory to save background noise related images'''
        dirName = path_outputs + "SII_BackgroundNoise"
        if not os.path.exists(dirName):
            os.mkdir(dirName)
        '''Defines the figure'''
        fig, ((ax0, ax1), (ax2, ax3)) = plt.subplots(2, 2)
        mz = float(mz)
        
        '''Single ion image figure'''
        ratio_sn = sn_tissue / sn_background
        diff_sn = sn_tissue - sn_background

        im = ax0.imshow(image, cmap=cm.YlGnBu_r)
        divider = make_axes_locatable(ax0)
        cax = divider.append_axes("right", size="5%", pad=0.1)
        plt.colorbar(im, cax=cax)
        ax0.title.set_text("S/N tissue:" + str(round(sn_tissue))
                           + "\n S/N bgr:" + str(round(sn_background))
                           + "\n ratio_SN:" + str(round(ratio_sn, 3))
                           + "\n diff_SN:" + str(round(diff_sn, 3)))

        '''Plot peak associated with image'''
        self.plot_peak_spectrum(ax1, data_spectrum, mz, mzlow, mzhigh)
        '''Plot binned image'''
        self.plot_sii_binned(ax2, image, npixels=3)
        '''Plot noise estimation'''
        self.plot_estimated_background(ax3, noise_sii_background, a, mask)

        plt.subplots_adjust(left=0.1, bottom=None, right=None, top=None, wspace=0.3, hspace=None)
        fig.set_size_inches(14, 8)
        fig.suptitle(str(round(mz, 3)) + "\n" + str(round(mzlow, 3)) + "-" + str(round(mzhigh, 3)))
        plt.savefig(path_outputs + "/SII_BackgroundNoise/" + str(round(mz, 3)) + ".png", dpi=100)
        #plt.show()
        plt.clf()
        plt.close('all')

    def plot_peak_spectrum(self, ax1, data_spectrum, mz, mzlow, mzhigh):
        """
        Function to plot peak of interest

        :param ax1: plot reference
        :type ax1: figure
        :param data_spectrum: mean spectrum to select the peak
        :type data_spectrum: DataSpectrum
        :param mz: peak centroid
        :type mz: float
        :param mzlow: peak's left minima
        :type mzlow: float
        :param mzhigh: peak's right minima
        :type mzhigh: float
        :return: plot of the peak
        :rtype: figure
        """

        intensity_window = data_spectrum.intensities[
            np.where((data_spectrum.mzs > mzlow) & (data_spectrum.mzs < mzhigh))]
        peak = np.max(intensity_window) + 2
        ax1.plot(data_spectrum.mzs, data_spectrum.intensities, color="black")
        ax1.set_xlim((mzlow - 0.05, mzhigh + 0.05))
        ax1.set_ylim((-0.1, peak + peak * 0.02))

        '''limits used to create the image'''
        ax1.axvline(mzlow, color="green")
        ax1.axvline(mzhigh, color="green")
        '''centroid'''
        ax1.axvline(mz, color="red")
        ax1.scatter(mz, np.max(intensity_window), color="green", marker="o")
        ax1.title.set_text("Mean spectrum")
        ax1.set_xlabel("mz")
        ax1.set_ylabel("Intensity")
        return ax1

    def plot_sii_binned(self, ax2, image, npixels=3):
        """
        Function to create a binned image. Useful for low intensity peaks

        :param ax2: plot reference
        :type ax2: figure
        :param image: image to binned
        :type image: array
        :param npixels: number of pixels to bin
        :type npixels: int
        :return: binned image
        :rtype: figure
        """

        dim0 = int(np.ceil(image.shape[0] / npixels))
        dim1 = int(np.ceil(image.shape[1] / npixels))
        im_bin = np.zeros((dim0, dim1))
        for i in range(dim0 - 1):
            for j in range(dim1 - 1):
                im_bin[i, j] = np.mean(image[npixels * i: npixels * (i + 1), npixels * j: npixels * (j + 1)])
            # fill in the edges
            for ii in range(dim0 - 1):
                im_bin[i, dim1 - 1] = np.mean(image[npixels * ii: npixels * (ii + 1), npixels * (dim1 - 1)::])
            for j in range(dim1 - 1):
                im_bin[dim0 - 1, j] = np.mean(image[npixels * (dim0 - 1)::, npixels * j: npixels * (j + 1)])
            # fill the final corner
            im_bin[dim0 - 1, dim1 - 1] = np.mean(image[npixels * (dim0 - 1)::, npixels * (dim1 - 1)::])

        im = ax2.imshow(im_bin, cmap=cm.YlGnBu_r)
        divider = make_axes_locatable(ax2)
        cax = divider.append_axes("right", size="5%", pad=0.1)
        plt.colorbar(im, cax=cax)
        ax2.title.set_text("Binned: " + str(npixels) + " pixels")
        return ax2

    def plot_estimated_background(self, ax3, noise_sii_background, a, mask):
        """
        Function to plot the calculated and extrapolated noise values

        :param ax3: image reference
        :type ax3: figure
        :param noise_sii_background: noise background
        :type noise_sii_background: numpy array
        :param a: estimated noise value
        :type a: float
        :param mask: mask of the tissue only region (to know number of pixels)
        :type mask: array
        :return: plot with estimated noise value for specific number of pixels
        :rtype: figure
        """

        ax3.scatter(np.log10(noise_sii_background[:, 0]),
                    np.log10(noise_sii_background[:, 1]), color="red", label="Measured")
        ax3.scatter(np.log10(len(mask[np.where(mask == 0)])), a, color="blue", label="Estimated")
        ax3.set_title("Background S/N", fontsize=20)
        ax3.set_xlabel("log10(N-pixels)", fontsize=16)
        ax3.set_ylabel("log10(std)", fontsize=16)
        ax3.tick_params(axis="both", which="minor", labelsize=14)
        ax3.locator_params(axis="y", nbins=4)
        ax3.locator_params(axis="x", nbins=4)
        ax3.legend()
        return ax3

    def plot_one_gaussian(self, xaxis, intensities, res, path_outputs):
        """
        Function to plot one gaussian fitting

        :param xaxis: x-axis range
        :type xaxis: array
        :param intensities: intensities in x-axis range
        :type intensities: array
        :param res: minimisation results
        :type res: class 'scipy.optimize._optimize.OptimizeResult'>
        :param path_outputs: path where to save the images
        :type path_outputs: str
        :return: None
        :rtype: None
        """

        plt.plot(xaxis, intensities, color="black", label="spectrum")
        plt.plot(xaxis[np.where(intensities == np.max(intensities))], np.max(intensities), "o", color="red")
        plt.plot(xaxis, self.gaussian_def.gaussian(res.x[0], res.x[1], res.x[2], xaxis),
            color="lawngreen", label="gaussian fit")
        plt.savefig(path_outputs + "Gaussian_fitting_most_intense_peaks/" + str(round(res.x[0], 3)) + "_simple.png")
        # plt.show()
        plt.clf()

    def plot_one_gaussian_separated(self, xaxis, intensities, xaxis_local, intensities_local, res, path_outputs):
        """
        Function to plot one gaussian fitting when there are more than one peak in the window

        :param xaxis: x-axis range
        :type xaxis: array
        :param intensities: intensities in x-axis range
        :type intensities: array
        :param xaxis_local: centroid value peak
        :type xaxis_local: float
        :param intensities_local: intensities value peak
        :type intensities_local: float
        :param res: minimisation results
        :type res: class 'scipy.optimize._optimize.OptimizeResult'>
        :param path_outputs: path where to save the images
        :type path_outputs: str
        :return: None
        :rtype: None
        """

        plt.plot(xaxis_local, intensities_local, "o", color="red")
        plt.plot(xaxis[np.where(abs(xaxis - xaxis_local) < 0.2)], intensities[np.where(abs(xaxis - xaxis_local) < 0.2)],
                 color="black", label="spectrum")
        plt.plot(xaxis[np.where(abs(xaxis - xaxis_local) < 0.2)],
                 self.gaussian_def.gaussian(res.x[0],
                                            res.x[1],
                                            res.x[2],
                                            xaxis[np.where(abs(xaxis - xaxis_local) < 0.2)]),
                 color="lawngreen", label="gaussian fit")
        plt.savefig(path_outputs + "Gaussian_fitting_most_intense_peaks/" + str(round(res.x[0], 3)) + "_simple.png")
        plt.clf()

    def plot_two_gaussians(self, xaxis, intensities, xaxis_local_1, intensities_local_1,
                           xaxis_local_2, intensities_local_2, res, path_outputs):
        """
        Function to plot two gaussian together

        :param xaxis: x-axis range
        :type xaxis: array
        :param intensities: intensities in x-axis range
        :type intensities: array
        :param xaxis_local_1: centroid value for first peak
        :type xaxis_local_1: float
        :param intensities_local_1: intensity value first peak
        :type intensities_local_1: float
        :param xaxis_local_2: centroid value for second peak
        :type xaxis_local_2: float
        :param intensities_local_2: intensities value second peak
        :type intensities_local_2: float
        :param res: minimisation results
        :type res: class 'scipy.optimize._optimize.OptimizeResult'>
        :param path_outputs: path where to save the images
        :type path_outputs: str
        :return: None
        :rtype: None
        """

        plt.plot(xaxis_local_1, intensities_local_1, "o", color="red")
        plt.plot(xaxis_local_2, intensities_local_2, "o", color="red")
        plt.plot(xaxis, intensities, color="black", alpha=0.8)
        plt.plot(xaxis,
                 self.gaussian_def.multigaussian(res.x[0], res.x[1], res.x[2], res.x[3], res.x[4], res.x[5], xaxis),
                 color="lawngreen")
        plt.plot(xaxis,
                 self.gaussian_def.gaussian(res.x[0], res.x[1], res.x[2], xaxis),
                 color="magenta", linestyle="dashed", alpha=0.5)
        plt.plot(xaxis,
                 self.gaussian_def.gaussian(res.x[3], res.x[4], res.x[5], xaxis),
                 color="orange", linestyle="dashed", alpha=0.5)
        plt.savefig(path_outputs + "Gaussian_fitting_most_intense_peaks/" + str(round(res.x[0], 3)) + "_multiple.png")
        plt.clf()

    def plot_three_gausssians(self, xaxis, intensities, xaxis_local_1, intensities_local_1,
                              xaxis_local_2, intensities_local_2,
                              xaxis_local_3, intensities_local_3,
                              res, path_outputs):
        """
        Function to plot three gaussian together

        :param xaxis: x-axis range
        :type xaxis: array
        :param intensities: intensities in x-axis range
        :type intensities: array
        :param xaxis_local_1: centroid value first peak
        :type xaxis_local_1: float
        :param intensities_local_1: intensities value first peak
        :type intensities_local_1: float
        :param xaxis_local_2: centroid value for second peak
        :type xaxis_local_2: float
        :param intensities_local_2: intensity value second peak
        :type intensities_local_2: float
        :param xaxis_local_3: centroid value third peak
        :type xaxis_local_3: flat
        :param intensities_local_3: intensities value third peak
        :type intensities_local_3: float
        :param res: minimisation results
        :type res: class 'scipy.optimize._optimize.OptimizeResult'>
        :param path_outputs: path where to save the images
        :type path_outputs: str
        :return: None
        :rtype: None
        """

        xaxis_fit_1 = xaxis[np.where(abs(xaxis - xaxis_local_1) < 0.2)]
        xaxis_fit_2 = xaxis[np.where(abs(xaxis - xaxis_local_2) < 0.2)]
        xaxis_fit_3 = xaxis[np.where(abs(xaxis - xaxis_local_3) < 0.2)]

        plt.plot(xaxis_local_1, intensities_local_1, "o", color="red")
        plt.plot(xaxis_local_2, intensities_local_2, "o", color="red")
        plt.plot(xaxis_local_3, intensities_local_3, "o", color="red")
        plt.plot(xaxis, intensities, color="black")
        plt.plot(xaxis,
                 self.gaussian_def.multigaussian_3g(res.x[0], res.x[1], res.x[2], res.x[3], res.x[4], res.x[5],
                                                    res.x[6], res.x[7], res.x[8], xaxis), color="lawngreen")
        plt.plot(xaxis_fit_1, self.gaussian_def.gaussian(res.x[0], res.x[1], res.x[2], xaxis_fit_1),
                 color="magenta", linestyle="dashed", alpha=0.5)
        plt.plot(xaxis_fit_2, self.gaussian_def.gaussian(res.x[3], res.x[4], res.x[5], xaxis_fit_2),
                 color="orange", linestyle="dashed", alpha=0.5)
        plt.plot(xaxis_fit_3, self.gaussian_def.gaussian(res.x[6], res.x[7], res.x[8], xaxis_fit_3),
                 color="lime", linestyle="dashed", alpha=0.5)
        plt.savefig(path_outputs + "Gaussian_fitting_most_intense_peaks/" + str(round(res.x[0], 3)) + "_multiple.png")
        plt.clf()

    def plot_one_gaussian_width_constraint(self, xaxis, intensities, local_peak, res, path_outputs):
        """
        Function to plot one gaussian fitting with constrains

        :param xaxis: x-axis range
        :type xaxis: array
        :param intensities: intensities in x-axis range
        :type intensities: array
        :param local_peak: centroid peak value
        :type local_peak: float
        :param res: minimisation results
        :type res: class 'scipy.optimize._optimize.OptimizeResult'>
        :param path_outputs: path where to save the images
        :type path_outputs: str
        :return: None
        :rtype: None
        """

        plt.plot(xaxis, intensities, color="black", label="spectrum")
        plt.plot(xaxis[np.where(xaxis == local_peak)], intensities[np.where(xaxis == local_peak)], "o", color="red")
        plt.plot(xaxis, self.gaussian_def.gaussian(res.x[0], res.x[1], res.x[2], xaxis),
            color="lawngreen", label="gaussian fit")
        plt.savefig(path_outputs + "Gaussian_fitting_width_constrained/" + str(round(res.x[0], 3)) + "_single.png")
        plt.clf()

    def plot_one_gaussian_separated_width_constraint(self, xaxis, intensities, xaxis_local, peak_local, res, func, std,
                                                     path_outputs):
        """
        Function to plot one gaussian fitting with constrains

        :param xaxis: x-axis range
        :type xaxis: array
        :param intensities: intensities in x-axis range
        :type intensities: array
        :param xaxis_local: centroid peak value
        :type xaxis_local: float
        :param peak_local: intensity peak value
        :type peak_local: float
        :param res: minimisation results
        :type res: class 'scipy.optimize._optimize.OptimizeResult'>
        :param func: correlation function
        :type func: poly1d
        :param std: new standard deviation
        :type std: float
        :param path_outputs: path where to save the images
        :type path_outputs: str
        :return: None
        :rtype: None
        """

        plt.plot(xaxis, intensities, color="black", label="spectrum")
        plt.plot(xaxis_local, peak_local, "o", color="red")
        # xaxis_1_fit = xaxis[(xaxis > xaxis_local - func(xaxis_local) - 6 * std)
        #                     & (xaxis < xaxis_local + func(xaxis_local) + 6 * std)]

        plt.plot(xaxis, self.gaussian_def.gaussian(res.x[0], res.x[1], res.x[2], xaxis),
            color="lawngreen", label="gaussian fit")
        plt.savefig(path_outputs + "Gaussian_fitting_width_constrained/" + str(round(res.x[0], 3)) + "_single.png")
        # plt.show()
        plt.clf()

    def plot_two_gaussians_width_constraint(self, xaxis, intensities, xaxis_local_1, peak_local_1, xaxis_local_2,
                                            peak_local_2, func, std, res, path_outputs):
        """
        Function to plot two gaussian with constrains

        :param xaxis: x-axis range
        :type xaxis: array
        :param intensities: intensities in x-axis range
        :type intensities: array
        :param xaxis_local_1: centroid value for first peak
        :type xaxis_local_1: float
        :param peak_local_1: intensity peak value first peak
        :type peak_local_1: float
        :param xaxis_local_2: centroid value for second peak
        :type xaxis_local_2: float
        :param peak_local_2: intensity peak value second peak
        :type peak_local_2: float
        :param func: correlation function
        :type func: poly1d
        :param std: new standard deviation
        :type std: float
        :param res: minimisation results
        :type res: class 'scipy.optimize._optimize.OptimizeResult'>
        :param path_outputs: path where to save the images
        :type path_outputs: str
        :return: None
        :rtype: None
        """

        plt.plot(xaxis_local_1, peak_local_1, "o", color="red")
        plt.plot(xaxis_local_2, peak_local_2, "o", color="red")
        plt.plot(xaxis, intensities, color="black", alpha=0.8)
        # xaxis_fit = xaxis[(xaxis > xaxis_local_1 - func(xaxis_local_1) - 6 * std)
        #                   & (xaxis < xaxis_local_2 + func(xaxis_local_2) + 6 * std)]

        plt.plot(xaxis, self.gaussian_def.multigaussian(
            res.x[0], res.x[1], res.x[2], res.x[3], res.x[4], res.x[5], xaxis),
                 color="lawngreen")

        xaxis_1_fit = xaxis[(xaxis > xaxis_local_1 - func(xaxis_local_1) - 6 * std)
                            & (xaxis < xaxis_local_1 + func(xaxis_local_1) + 6 * std)]
        xaxis_2_fit = xaxis[(xaxis > xaxis_local_2 - func(xaxis_local_2) - 6 * std)
                            & (xaxis < xaxis_local_2 + func(xaxis_local_2) + 6 * std)]

        plt.plot(xaxis_1_fit, self.gaussian_def.gaussian(res.x[0], res.x[1], res.x[2], xaxis_1_fit),
                 color="magenta", linestyle="dashed", alpha=0.5)
        plt.plot(xaxis_2_fit, self.gaussian_def.gaussian(res.x[3], res.x[4], res.x[5], xaxis_2_fit),
                 color="orange", linestyle="dashed", alpha=0.5)
        plt.savefig(path_outputs + "Gaussian_fitting_width_constrained/" + str(round(res.x[0], 3)) + "_double.png")
        # plt.show()
        plt.clf()

    def plot_three_gausssians_width_constraint(self, xaxis, intensities,
                                               xaxis_local_1, peak_local_1,
                                               xaxis_local_2, peak_local_2,
                                               xaxis_local_3, peak_local_3, func, std, res, path_outputs):
        """
        Function to plot three gaussian with constrains

        :param xaxis: x-axis range
        :type xaxis: array
        :param intensities: intensities in x-axis range
        :type intensities: array
        :param xaxis_local_1: centroid value for first peak
        :type xaxis_local_1: float
        :param peak_local_1: intensity peak value first peak
        :type peak_local_1: float
        :param xaxis_local_2: centroid value for second peak
        :type xaxis_local_2: float
        :param peak_local_2: intensity peak value second peak
        :type peak_local_2: float
        :param xaxis_local_3: centroid value for third peak
        :type xaxis_local_3: float
        :param peak_local_3: intensity peak value third peak
        :type peak_local_3: float
        :param func: correlation function
        :type func: poly1d
        :param std: new standard deviation
        :type std: float
        :param res: minimisation results
        :type res: class 'scipy.optimize._optimize.OptimizeResult'>
        :param path_outputs: path where to save the images
        :type path_outputs: str
        :return: None
        :rtype: None
        :return:
        """

        xaxis_fit_1 = xaxis[(xaxis > xaxis_local_1 - func(xaxis_local_1) - 6 * std)
                            & (xaxis < xaxis_local_1 + func(xaxis_local_1) + 6 * std)]

        xaxis_fit_2 = xaxis[(xaxis > xaxis_local_2 - func(xaxis_local_2) - 6 * std)
                            & (xaxis < xaxis_local_2 + func(xaxis_local_2) + 6 * std)]

        xaxis_fit_3 = xaxis[(xaxis > xaxis_local_3 - func(xaxis_local_3) - 6 * std)
                            & (xaxis < xaxis_local_3 + func(xaxis_local_3) + 6 * std)]

        plt.plot(xaxis_local_1, peak_local_1, "o", color="red")
        plt.plot(xaxis_local_2, peak_local_2, "o", color="red")
        plt.plot(xaxis_local_3, peak_local_3, "o", color="red")
        plt.plot(xaxis, intensities, color="black")
        plt.plot(xaxis, self.gaussian_def.multigaussian_3g(res.x[0], res.x[1], res.x[2], res.x[3], res.x[4],
                                                           res.x[5], res.x[6], res.x[7], res.x[8], xaxis),
                 color="lawngreen")
        plt.plot(xaxis_fit_1, self.gaussian_def.gaussian(res.x[0], res.x[1], res.x[2], xaxis_fit_1),
                 color="magenta", linestyle="dashed", alpha=0.5)
        plt.plot(xaxis_fit_2, self.gaussian_def.gaussian(res.x[3], res.x[4], res.x[5], xaxis_fit_2),
                 color="orange", linestyle="dashed", alpha=0.5)
        plt.plot(xaxis_fit_3, self.gaussian_def.gaussian(res.x[6], res.x[7], res.x[8], xaxis_fit_3),
                 color="purple", linestyle="dashed", alpha=0.5)
        plt.savefig(path_outputs + "Gaussian_fitting_width_constrained/" + str(round(res.x[0], 3)) + "_triple.png")
        plt.clf()



"""Main file to be run in python to apply the pre-processing in a MSI dataset.
Several tasks can be done with nectar-msi package. Here we show an example of the most
standard options used to obtain a final list of peaks for further analysis."""

"""___Built-In Modules___"""
from nectar_msi import Readers, Savers, DataOperations, NoiseCorrection, PeakPicking, DatabaseMatching, Plotting

"""___Third-Party Modules___"""
import sys


"""___Authorship___"""
__author__ = "Ariadna Gonzalez"
__created__ = "07/09/2023"
__maintainer__ = "Ariadna Gonzalez"
__email__ = "ariadna.gonzalez@npl.co.uk"

if not sys.warnoptions:
    import warnings

    warnings.simplefilter("ignore")


def main():
    reader = Readers()
    saver = Savers()
    dataop = DataOperations()
    noisecorrection = NoiseCorrection()
    databasematching = DatabaseMatching()
    peakpicking = PeakPicking()
    plotting = Plotting()

    '''Define paths'''
    path_data = "..."
    file = "example.imzML"
    path_outputs = "..."
    polarity = "positive"  # in case you want to cross-match with HMDB to calculate the adducts
    modality = "maldi"  # example of used modality

    '''Read imzML data'''
    data = reader.read_imzml(path_data + file)

    '''Create full mean spectrum'''
    total_mean = dataop.get_mean_spectrum(data)
    saver.save_spectrum_hdf5(path_outputs + 'total_mean.hdf5', total_mean)

    '''Separate tissue from background'''
    data_masked = dataop.background_subtraction(data, total_mean, path_outputs, n_clusters=2, show_plot=True)
    saver.save_hdf5(path_outputs + 'example_masked.hdf5', data_masked)
    # saver.save_imzML(path_outputs + 'example_masked.imzML', data_masked)  # to save in imzML format

    # data_masked = reader.read_hdf5(path_outputs + 'example_masked.hdf5')  # to read the masked file in hdf5 format
    # data_masked = reader.read_imzML(path_outputs + 'example_masked.imzML')  # to read the masked file in imzML format

    '''Creates mean spectra for tissue and background'''
    mean_tissue, mean_background = dataop.get_mean_spectrum_tissue_background(data_masked, mean_tissue=True,
                                                                              mean_background=True)
    saver.save_spectrum_hdf5(path_outputs + 'mean_tissue.hdf5', mean_tissue)
    saver.save_spectrum_hdf5(path_outputs + 'mean_background.hdf5', mean_background)

    # mean_tissue = reader.read_hdf5_spectrum(path_outputs + 'mean_tissue.hdf5')  # to read the spectrum
    # mean_background = reader.read_hdf5_spectrum(path_outputs + 'mean_background.hdf5')

    '''Apply noise correction to spectrum'''
    mean_tissue_corrected = noisecorrection.noise_correction_with_chemical_noise(mean_tissue,
                                                                                 plot_noise=True,
                                                                                 plot_chemicalnoise=True)
    mean_tissue_corrected = noisecorrection.noise_correction(mean_tissue, plot_noise=True)  # If yor data does not have
    # chemical noise (sinusoidal noise), you can determine the signal/noise with the SigmaClipping function

    saver.save_spectrum_hdf5(path_outputs + 'mean_tissue_corrected.hdf5', mean_tissue_corrected)
    # mean_tissue_corrected = reader.read_hdf5_spectrum(path_outputs + 'mean_tissue_corrected.hdf5')

    '''Peak picking and background contamination removal'''
    list_of_fittings, full_list_of_fittings = peakpicking.peak_picking(mean_tissue_corrected, path_outputs,
                                                                       plot_peaks=True,
                                                                       save_tables=True,
                                                                       save_fitting=True)

    # full_list_of_fittings = pd.read_csv("X:\\Ariadna\\PDAC\\MALDI\\nectar_outputs\\full_list_of_fittings.csv")

    peaks_classification = noisecorrection.backgroundnoise.backgroundnoise_identification(data_masked,
                                                                                          mean_tissue_corrected,
                                                                                          full_list_of_fittings,
                                                                                          path_outputs,
                                                                                          save_plot_backgroundNoise=True
                                                                                          )

    # peaks_classification = pd.read_csv("X:\\Ariadna\\PDAC\\MALDI\\nectar_outputs\\Peaks_classification.csv")

    '''Final selection of peaks according to S/N'''
    final_list = peaks_classification.loc[(peaks_classification['diff[Tis/bak] S/N'] >= 0) &
                                          (peaks_classification['ratio[Tis/bak] S/N'] >= 5)]

    #final_list.to_csv("X:\\Ariadna\\PDAC\\MALDI\\nectar_outputs\\Compounds_of_interest_final_list.csv", index=False)

    '''Save reduced data cubes'''
    saver.save_final_DataCube(data_masked, final_list, path_outputs, save_imzml=True, save_hdf5=False, save_mat=False)

    '''HMDB database cross matching'''
    databasematching.database_matching(final_list, polarity, modality, path_outputs, ppm=30.)

    '''Plot of selected peaks'''
    plotting.plot_sii_final_list(data_masked, mean_tissue_corrected, final_list, path_outputs, save_fig=True)

    print('Finished!')

if __name__ == "__main__":
    main()



"""Class to correct from background noise/contamination from matrix/solvent"""

"""___Built-In Modules___"""
from nectar_msi.plotting.plotting import Plotting

"""___Third-Party Modules___"""
import numpy as np
from tqdm import tqdm

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class BackgroundNoise:
    def __init__(self):
        self.plotting = Plotting()
        pass

    def backgroundnoise_identification(self, data_ori, data_spectrum, list_of_peaks, path_outputs,
                                       save_plot_backgroundNoise=False):
        """
        Function to identified and remove from the final list of peaks the background noise peaks
        :param data_ori: MSI datacube
        :type data_ori: DataCube
        :param data_spectrum: mean spectrum
        :type data_spectrum: DataSpectrum
        :param list_of_peaks: First de-noised list of peaks
        :type list_of_peaks: pandas DataFrame
        :param path_outputs: outputs path
        :type path_outputs: str
        :param save_plot_backgroundNoise: option to save the single ion images and the estimated S/N
        :type save_plot_backgroundNoise: bool
        :return: list_of_peaks (pandas DataFrame)
        """

        ''' The input DataCube needs to have tissue and background mask to be able to use this function'''
        mask = data_ori.pixelmask

        if len(mask) == (mask == 0).sum():
            print("!Error: Your data does not have a mask to apply the background correction. \n"
                  'Please, apply the "background_subtraction" function first.')
            exit()

        ''' Select what cluster number corresponds to tissue and background'''
        cluster_tissue = int(input("Please enter the tissue cluster number: "))
        cluster_background = int(input("Please enter the background cluster number: "))
        mask_2d = np.reshape(mask, (data_ori.shape[1], data_ori.shape[0])).T
        print("Number of background pixels", len(mask_2d[np.where(mask_2d == cluster_tissue)]))
        print("Number of tissue pixels", len(mask_2d[np.where(mask_2d == cluster_background)]))

        print("Background noise correction running...")
        '''Including new columns in the dataframe for the final list'''
        list_of_peaks["meanIntTotal"] = ""
        list_of_peaks["meanTissue"] = ""
        list_of_peaks["meanBackground"] = ""
        list_of_peaks["log2(meanTissue/meanBackground)"] = ""
        list_of_peaks["Tissue S/N"] = ""
        list_of_peaks["Background S/N"] = ""
        list_of_peaks["ratio[Tis/bak] S/N"] = ""
        list_of_peaks["diff[Tis/bak] S/N"] = ""
        list_of_peaks["left min"] = list_of_peaks["left min"].astype(float)
        list_of_peaks["right min"] = list_of_peaks["right min"].astype(float)

        '''Calculate the S/N for background and tissue'''
        for i in tqdm(range(len(list_of_peaks["meas mz"]))):
            image = data_ori.get_intensities_between_mzs(list_of_peaks["left min"][i], list_of_peaks["right min"][i])
            background = image[np.where(mask_2d == cluster_background)]
            background_sum = np.mean(background) * len(mask_2d[np.where(mask_2d == cluster_background)])
            background_std = np.std(background) * np.sqrt(len(mask_2d[np.where(mask_2d == cluster_background)]))
            sn_backgroundpixels = background_sum / background_std
            signal_tissue = np.sum(image[np.where(mask_2d == cluster_tissue)])

            noise_sii_background = np.zeros((6, 2))
            if not np.isfinite(sn_backgroundpixels):
                print("finite")
                continue

            '''These values are specific for the case study we were studying. Estimation of the background noise
            for the number of pixels on tissue'''
            list = [1, 5, 10, 50, 100, 250]
            list_n = [3000, 1000, 500, 100, 50, 20]
            for j, pixel in enumerate(list):
                pixel_noise_average_bg = np.zeros(list_n[j])
                for jj in range(0, list_n[j], 1):
                    pixel_noise_bg = np.random.choice(image[np.where(mask_2d == cluster_background)],
                                                      pixel, replace=False)
                    pixel_noise_average_bg[jj] = np.sum(pixel_noise_bg)

                noise_sii_background[j] = [pixel, np.std(pixel_noise_average_bg)]

            a = np.polyfit(np.log10(noise_sii_background[:, 0]).astype(np.float32),
                           np.log10(noise_sii_background[:, 1]).astype(np.float32), 1)
            func_all_peaks = np.poly1d(a)
            a = func_all_peaks(np.log10(len(mask_2d[np.where(mask_2d == cluster_tissue)])))
            bckgr_noise = 10**a

            '''Calculation of S/N, ration and difference of tissue and background'''
            sn_tissue = signal_tissue / bckgr_noise
            sn_background = background_sum / bckgr_noise
            ratio_sn = sn_tissue / sn_background
            diff_sn = sn_tissue - sn_background

            if save_plot_backgroundNoise == True:
                self.plotting.plot_backgroundNoise(image, data_spectrum,
                                                   list_of_peaks["meas mz"][i],
                                                   list_of_peaks["left min"][i],
                                                   list_of_peaks["right min"][i],
                                                   noise_sii_background, a, mask_2d,
                                                   sn_tissue, sn_background, path_outputs)

            '''Filling the final table'''
            image = image.flatten()
            mean_intensities = np.mean(image)
            mean_tissue = np.sum(image[np.where(mask == cluster_tissue)]) / len(image[np.where(mask == cluster_tissue)])
            mean_background = np.sum(image[np.where(mask == cluster_background)]) / \
                              len(image[np.where(mask == cluster_background)])

            list_of_peaks["meanIntTotal"][i] = mean_intensities
            list_of_peaks["meanTissue"][i] = mean_tissue
            list_of_peaks["meanBackground"][i] = mean_background
            list_of_peaks["log2(meanTissue/meanBackground)"][i] = np.log2(mean_tissue / mean_background)
            list_of_peaks["Tissue S/N"][i] = sn_tissue
            list_of_peaks["Background S/N"][i] = sn_background
            list_of_peaks["ratio[Tis/bak] S/N"][i] = ratio_sn
            list_of_peaks["diff[Tis/bak] S/N"][i] = diff_sn

        list_of_peaks.to_csv(path_outputs + "peaks_classification.csv", index=False)
        print("!Final list of peaks saved in: " + path_outputs)

        return list_of_peaks



"""Class for performing operations on the data"""

"""___Built-In Modules___"""
from nectar_msi.data_formats.data_spectrum import DataSpectrum
from nectar_msi.data_formats.data_cube import DataCube
from nectar_msi.noise_correction.noise_determination import NoiseDetermination
from nectar_msi.data_processing.savers import Savers

"""___Third-Party Modules___"""
import numpy as np
from scipy import stats as st
import matplotlib.pyplot as plt
import h5py
import matplotlib.cm as cm
from sklearn.cluster import KMeans
from tqdm import tqdm
import os

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class DataOperations:
    def __init__(self):
        self.noisedetermination = NoiseDetermination()
        self.saver = Savers()
        pass

    def get_mean_spectrum(self, data_cube, start_mz=None, end_mz=None, xaxis=None, scaling_power=0.5):
        """
        Function to obtain the mean spectrum of a data_cube without assigned mask

        :param data_cube: data_cube of interest
        :type data_cube: DataCube
        :param start_mz: first value in the x axis of the mean spectrum (default to None)
        :type start_mz: float
        :param end_mz: last value in the x axis of the mean spectrum (default to None)
        :type end_mz: float
        :param xaxis: x axis of the spectrum (default to None, in which case is created)
        :type xaxis: array
        :param scaling_power: the difference in m/z between two neighbouring channels mz_1 and mz_2 is given by
        mz2-mz_1 = a * mz_1**p, where a is the scaling factor and p is the scaling power (default = 0.5)
        :type scaling_power: float
        :return: mean spectrum
        :rtype: DataSpectrum
        """

        # Creates common x-axis for all pixels
        if xaxis is None:
            xaxis, scaling_factor = self.make_xaxis(data_cube, start_mz=start_mz, end_mz=end_mz,
                                                    scaling_power=scaling_power)

        print("Creating mean spectrum...")
        mean_intensities = np.zeros_like(xaxis)
        n_valid_spectra = 0  # only selects pixels with data on it
        for index in tqdm(range(len(data_cube.idxs))):
            spectrum = data_cube.get_spectrum_idx(data_cube.idxs[index])
            if spectrum is not None:
                intensities_pixeli = self.process_pixel_intensities(spectrum, xaxis)
                mean_intensities += intensities_pixeli
                n_valid_spectra += 1
        mean_intensities = mean_intensities / n_valid_spectra  # mean intensities
        spectrum = DataSpectrum(-99, [-99, -99], xaxis, mean_intensities)  # mean spectrum as a DataSpectrum object
        print("Mean spectrum created!")
        return spectrum

    def get_mean_spectrum_tissue_background(self, data_cube, start_mz=None, end_mz=None, xaxis=None, mean_tissue=True,
                                            mean_background=False, scaling_power=0.5):

        """
        Function to create the mean spectrum of the tissue and/or background based in a
        previously set mask

        :param data_cube: data_cube from where we want to calculate the mean spectra
        :type data_cube: DataCube
        :param start_mz: first value in the x axis of the mean spectrum (default to None)
        :type start_mz: float
        :param end_mz: last value in the x axis of the mean spectrum (default to None)
        :type end_mz: float
        :param xaxis: x axis of the spectrum (default to None, in which case is created)
        :type xaxis: array
        :param mean_tissue: Option to choose if we want to create the mean spectrum
        of the tissue (default True)
        :type mean_tissue: bool
        :param mean_background: Option to choose if we want to create the mean spectrum
        of the background (default False)
        :type mean_background: bool
        :param scaling_power: the difference in m/z between two neighbouring channels mz_1 and mz_2 is given by
        mz2-mz_1 = a * mz_1**p, where a is the scaling factor and p is the scaling power (default = 0.5)
        :type scaling_power: float
        :return: spectrum for tissue and/or background
        :rtype: DataSpectrum
        """
        if xaxis is None:
            xaxis, scaling_factor = self.make_xaxis(data_cube,
                                                    start_mz=start_mz, end_mz=end_mz, scaling_power=scaling_power)

        cluster_tissue = int(input("Please enter the tissue cluster number: "))
        cluster_background = int(input("Please enter the background cluster number: "))

        if mean_tissue == True and mean_background == True:
            print("Creating mean spectrum from tissue...")
            n_valid_spectra = 0
            mean_intensities = np.zeros_like(xaxis)
            for index in tqdm(range(len(data_cube.idxs))):
                if data_cube.pixelmask[index] == cluster_tissue:
                    spectrum = data_cube.get_spectrum_idx(data_cube.idxs[index])
                    if spectrum is not None:
                        intensities_pixeli = self.process_pixel_intensities(spectrum, xaxis)
                        mean_intensities += intensities_pixeli
                        n_valid_spectra += 1
            mean_intensities = mean_intensities / n_valid_spectra
            spectrum_tissue = DataSpectrum(-99, [-99, -99], xaxis, mean_intensities)
            print("Tissue-spectrum created!")

            print("Creating mean spectrum from background...")
            n_valid_spectra = 0
            mean_intensities = np.zeros_like(xaxis)
            for index in tqdm(range(len(data_cube.idxs))):
                if data_cube.pixelmask[index] == cluster_background:
                    spectrum = data_cube.get_spectrum_idx(data_cube.idxs[index])
                    if spectrum is not None:
                        spectrum = data_cube.get_spectrum_idx(data_cube.idxs[index])
                        intensities_pixeli = self.process_pixel_intensities(spectrum, xaxis)
                        mean_intensities += intensities_pixeli
                        n_valid_spectra += 1
            mean_intensities = mean_intensities / n_valid_spectra
            spectrum_background = DataSpectrum(-99, [-99, -99], xaxis, mean_intensities)
            print("Background-spectrum created!")

            return spectrum_tissue, spectrum_background

        elif mean_background:
            print("Creating mean spectrum from background...")
            n_valid_spectra = 0
            mean_intensities = np.zeros_like(xaxis)
            for index in tqdm(range(len(data_cube.idxs))):
                if data_cube.pixelmask[index] == cluster_background:
                    spectrum = data_cube.get_spectrum_idx(data_cube.idxs[index])
                    if spectrum is not None:
                        spectrum = data_cube.get_spectrum_idx(data_cube.idxs[index])
                        intensities_pixeli = self.process_pixel_intensities(spectrum, xaxis)
                        mean_intensities += intensities_pixeli
                        n_valid_spectra += 1
            mean_intensities = mean_intensities / n_valid_spectra
            spectrum_background = DataSpectrum(-99, [-99, -99], xaxis, mean_intensities)
            print("Background-spectrum created!")

            return spectrum_background

        elif mean_tissue:
            print("Creating mean spectrum from tissue...")
            n_valid_spectra = 0
            mean_intensities = np.zeros_like(xaxis)
            for index in tqdm(range(len(data_cube.idxs))):
                if data_cube.pixelmask[index] == cluster_tissue:
                    spectrum = data_cube.get_spectrum_idx(data_cube.idxs[index])
                    if spectrum is not None:
                        intensities_pixeli = self.process_pixel_intensities(spectrum, xaxis)
                        mean_intensities += intensities_pixeli
                        n_valid_spectra += 1
            mean_intensities = mean_intensities / n_valid_spectra
            spectrum_tissue = DataSpectrum(-99, [-99, -99], xaxis, mean_intensities)
            print("Tissue-spectrum created!")

            return spectrum_tissue

    def process_pixel_intensities(self, data_spectrum, xaxis):
        """
        Function to assign intensities of each pixel into the corresponding channel of the common x-axis

        :param data_spectrum: spectrum of each pixel
        :type data_spectrum: DataSpectrum
        :param xaxis: common x-axis for all pixels
        :type xaxis: array
        :return: Array of intensities with the same shape as x-axis
        :rtype: array
        """
        mzs, intensities = data_spectrum.mzs, data_spectrum.intensities
        yvals = np.zeros_like(xaxis)

        for i, mz in enumerate(mzs):
            yvals[np.argmin(np.abs(xaxis - mz))] = intensities[i]
        return yvals

    def make_xaxis(self, data_cube, start_mz=None, end_mz=None, scaling_power=0.5):
        """
        Function to create a common x-axis for all pixels

        :param data_cube: data cube of interest
        :type data_cube: DataCube
        :param start_mz: first m/z value of the spectrum
        :type start_mz: float
        :param end_mz: last m/z value of the spectrum
        :type end_mz: float
        :param scaling_power: the difference in m/z between two neighbouring channels mz_1 and mz_2 is given by
        mz2-mz_1 = a * mz_1**p, where a is the scaling factor and p is the scaling power (default = 0.5)
        :type scaling_power: float
        :return: new x axis for the spectrum and scaling factor
        :rtype: array, float
        """
        print("Creating x axis")
        indexmax = -99
        lenmax = 0
        # finds the spectrum with more measurements in the m/z range to calculate the scaling factor
        for index in range(len(data_cube.coord)):
            if data_cube.get_spectrum_idx(data_cube.idxs[index]) is not None:
                len_spectrum_per_pixel = len(data_cube.get_spectrum_idx(data_cube.idxs[index]).mzs)
                if len_spectrum_per_pixel > lenmax:
                    lenmax = len_spectrum_per_pixel
                    indexmax = index

        spectrum = data_cube.get_spectrum_idx(data_cube.idxs[indexmax])
        # Uses this spectrum as reference to calculate the scaling factor
        # scaling factor determines the channel size (bin) across the m/z range
        scaling_factor, scaling_std = self.find_scaling_factor_xaxis(spectrum.mzs, scaling_power=scaling_power)
        # print("scaling factor", scaling_factor)
        # Creates the x-axis for the mean spectrum
        new_mz = self.fill_xaxis(spectrum.mzs, scaling_factor, scaling_power, start_mz=start_mz, end_mz=end_mz)
        return new_mz, scaling_factor

    def find_scaling_factor_xaxis(self, xaxis, scaling_power=0.5):
        """
        Function to find the scaling factor to be used to create the x-axis for the mean spectrum

        :param xaxis: x-axis of the selected spectrum with most number of measurements
        :type xaxis: array
        :param scaling_power: the difference in m/z between two neighbouring channels mz_1 and mz_2 is given by
        mz2-mz_1 = a * mz_1**p, where a is the scaling factor and p is the scaling power (default=0.5)
        :type scaling_power: float
        :return: scaling factor and standard deviation of the scaling factor
        :rtype: floats
        """

        # Distances among all measurement in the spectrum
        scaling_candidates = (xaxis[1:] - xaxis[:-1]) / xaxis[1:] ** scaling_power
        # Sigma clipping to determine the most common value, and therefore select the scaling factor
        sigma_new, average, values = self.noisedetermination.SigmaClip(scaling_candidates)
        # print(scaling_power,np.round(scaling_candidates/average,1),st.mode(np.round(scaling_candidates/average,1))[0][0])
        if st.mode(np.round(scaling_candidates / average, 1))[0][0] != 1:
            print(
                "Are you sure the scaling power of your x-axis is correct? It looks like the data indicates it is not."
            )
        return average, sigma_new

    def fill_xaxis(self, xaxis, scaling_factor, scaling_power, start_mz=None, end_mz=None):
        """
        Function to create the common x-axis for all pixels

        :param xaxis: x-axis of the reference spectrum
        :type xaxis: array
        :param scaling_factor: the difference in m/z between two neighbouring channels mz_1 and mz_2 is given by
        mz2-mz_1 = a * mz_1**p, where a is the scaling factor and p is the scaling power
        :type scaling_factor: float
        :param scaling_power: the difference in m/z between two neighbouring channels mz_1 and mz_2 is given by
        mz2-mz_1 = a * mz_1**p, where a is the scaling factor and p is the scaling power
        :type scaling_power: float
        :param start_mz: first value in the x axis of the mean spectrum (default to None,
        we need to set this value if xaxis = None)
        :type start_mz: float
        :param end_mz: last value in the x axis of the mean spectrum (default to None)
        :type end_mz: float
        :return: new x-axis for mean spectrum
        :rtype: array
        """
        xaxis2 = np.zeros(5000000)
        iextra = 0
        i = 0
        if start_mz is None:
            start_mz = xaxis[0]
        if end_mz is None:
            end_mz = xaxis[-1]
        xaxis2[0] = start_mz
        # (xaxis[1:]-xaxis[:-1])/xaxis[1:]**1.5 =scaling
        while xaxis2[i] < end_mz and i - iextra + 1 < len(xaxis):
            # distance between previous xaxis point and next theoretical xaxis point (i.e. dist between x[i+1] and x[i])
            dist_theor = xaxis[i - iextra + 1] ** scaling_power * scaling_factor
            # distance between previous xaxis point and next measured xaxis point
            dist_measured = xaxis[i - iextra + 1] - xaxis2[i]
            if dist_measured / dist_theor > 1.5:
                iextra += 1
                xaxis2[i + 1] = xaxis2[i] + xaxis2[i] ** scaling_power * scaling_factor
            else:
                # print(xaxis[i-iextra+1])
                xaxis2[i + 1] = xaxis[i - iextra + 1]
            i += 1

        xaxis2 = xaxis2[np.where((xaxis2 >= start_mz) & (xaxis2 < end_mz))]
        return xaxis2

    def find_minima_around_peak(self, xaxis_window, spectrum_window, peak):
        """
        Function to find the minima of a given m/z value

        :param xaxis_window: Window size of mz values around where to look for the minima
        :type xaxis_window: array
        :param spectrum_window: Intensity values of the xaxis_window
        :type spectrum_window: array
        :param peak: centroid value
        :type peak: float
        :return: minima around the m/z value
        :rtype: floats
        """

        dydx = np.diff(spectrum_window) / np.diff(xaxis_window)
        right = [xaxis_window[-1]]
        left = [xaxis_window[0]]
        for ii in range(len(dydx) - 1):
            if np.sign(dydx[ii + 1]) > np.sign(dydx[ii]):
                local_mz = xaxis_window[ii + 1]
                if local_mz > peak:
                    right = np.append(right, xaxis_window[ii + 1])
                else:
                    left = np.append(left, xaxis_window[ii + 1])
        left_min = np.argmin(np.abs(left - peak))
        right_min = np.argmin(np.abs(right - peak))
        mzlow = left[left_min]
        mzhigh = right[right_min]
        return mzlow, mzhigh

    def background_subtraction(self, data_cube, data_mean, path_outputs, n_clusters=2, show_plot=False):
        """
        Function to separate between tissue and background by K-means

        :param data_cube: data_cube from which we want to separate tissue and background
        :type data_cube: DataCube
        :param data_mean: total mean spectrum (tissue+background)
        :type data_mean: DataSpectrum
        :param path_outputs: path to save the KMeans image
        :type path_outputs: str
        :param n_clusters: Number of clusters for k-means
        :type n_clusters: int
        :param show_plot: Option to show the KMeans resulted plot (default False)
        :type show_plot: bool
        :return: data_cube with a set mask separating tissue and background
        :rtype: DataCube
        """
        dirName = path_outputs
        if not os.path.exists(dirName):
            os.mkdir(dirName)

        print("Background subtraction...")
        index_intensities = np.argsort(data_mean.intensities[np.where(~np.isnan(data_mean.intensities))])[::-1][:500]
        peaks_mz = data_mean.mzs[index_intensities]
        peaks_mz.sort()

        cube = np.zeros((data_cube.shape[0] * data_cube.shape[1], len(peaks_mz)))
        for i in range(len(peaks_mz)):
            for j in range(data_cube.shape[0] * data_cube.shape[1]):
                spec = data_cube.get_spectrum_idx(data_cube.idxs[j])
                if spec is not None:
                    cube[j, i] = spec.get_intensity_between_mzs(
                        peaks_mz[i] * (1 - 30 * 10**-6),
                        peaks_mz[i] * (1 + 30 * 10**-6))

        print("Applying K-means...")
        cluster = KMeans(init="random", n_clusters=n_clusters, n_init="auto", max_iter=300).fit_predict(cube)
        cluster_reshape = cluster.reshape(data_cube.shape[1], data_cube.shape[0], 1)
        cluster_reshape = np.moveaxis(cluster_reshape, 0, 1)

        print("Setting tissue-background mask...")
        data_cube.set_pixel_mask_2D(cluster_reshape)

        plt.matshow(data_cube.get_pixel_mask_2D(), cmap=cm.get_cmap("RdBu", 2))
        plt.colorbar()
        if show_plot:
            plt.show()
        plt.savefig(path_outputs + "KMeans_tissue_background.png")
        print('Figure save in ' + path_outputs + 'KMeans_tissue_background.png')
        plt.clf()
        return data_cube

    def read_mean_from_sa(self, path_sa):
        """
        Function to read the spectrum created in spectral analysis
        :param path_sa: path to the files created by spectral analysis
        :type path_sa: string
        :return: spectrum
        :rtype: DataSpectrum
        """
        mz = h5py.File(path_sa+'totalSpectrum_mzvalues.mat')
        intensities = h5py.File(path_sa+'totalSpectrum_intensities.mat')

        mz = mz["totalSpectrum_mzvalues"]
        intensities = intensities["totalSpectrum_intensities"]
        mz = [item for sublist in mz for item in sublist]
        intensities = [item for sublist in intensities for item in sublist]
        mean = DataSpectrum(-99, [-99, -99], mz, intensities)
        print('Mean spectrum from Spectral Analysis created')
        return mean



import io
import os
import re

from setuptools import find_packages
from setuptools import setup
import versioneer

exec(open('nectar_msi/_version.py').read())

def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


setup(
    version=__version__,
    name="nectar_msi",
    url="https://gitlab.npl.co.uk/nice-msi/nectar",
    license="GNU Lesser General Public License v3.0",
    author="Ariadna Gonzalez-Fernandez",
    author_email="ariadna.gonzalez@npl.co.uk",
    description="NECTAR_MSI (NoisE CorrecTion AlgoRithm) a python package for noise determination and correction in MSI.",
    long_description=read("README.md"),
    packages=find_packages(exclude=("tests",)),
    install_requires=['numpy', 'matplotlib', 'pyimzml', 'h5py', 'scipy', 'pandas', 'scikit-learn', 'tqdm'],
    extras_require={"dev": ["pre-commit", "tox", "sphinx", "sphinx_rtd_theme", 'jupyter']},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



"""Class for creating and reading spectra"""

"""___Third-Party Modules___"""
import numpy as np

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class DataSpectrum:
    def __init__(self, idx, coordinates, mzs, intensities, mzmask=None, offsets=None, metadata={}):
        """
        DataSpectrum class initialiser

        :param idx: Id of the spectrum in the datacube
        :type idx: int
        :param coordinates: coordinates of the spectrum in the data_cube
        :type coordinates: tuple
        :param mzs: mz values in the spectrum
        :type mzs: array float64
        :param intensities: intensity values in the spectrum
        :type intensities: array float32
        :param mzmask: mask to mask part of the mz values (default to zeros)
        :type mzmask: array of int (optional)
        :param offsets: given reference mzs values, offsets to those references (default to empty)
        :type offsets: dict (optional)
        :param metadata:NOT DEVELOPED - Metadata of the spectrum (default to empty)
        :type metadata: dict (optional)
        """
        self.idx = idx
        self.coordinates = coordinates
        self.mzs = np.array(mzs, dtype="float64")
        self.intensities = np.array(intensities, dtype="float32")
        if mzmask is None:
            self.mzmask = np.zeros_like(mzs, dtype="u2")
        else:
            self.mzmask = np.array(mzmask, dtype="u2")
        if offsets is None:
            self.offsets = {}
        else:
            self.offsets = offsets
        self.spectrummetadata = metadata

    def set_mzs(self, mzs_new):
        """
        Function to set new mzs values in the spectrum

        :param mzs_new: new mzs values to set
        :type mzs_new: np.array
        :return: new mzs values in the spectrum
        :rtype: array of floats
        """
        self.mzs = np.array(mzs_new, dtype="float64")

    def set_intensities(self, new_intensities):
        """
        Funtion to set new intensity values in the spectrum

        :param new_intensities: new intensity values to set
        :type new_intensities: np.array
        :return: new intensity values in the spectrum
        :rtype: array of floats
        """
        self.intensities = np.array(new_intensities, dtype="float32")

    def set_mzmask(self, new_mask):
        """
        Function to set the mzmask in the spectrum class

        :param new_mask: mask to be included in the spectrum class
        :type new_mask: np.array
        :return: mask
        :rtype: array of int
        """
        self.mzmask = np.array(new_mask, dtype="u2")

    def set_offset(self, mz, offset):
        """
        Function to set the dictionary of offset values

        :param mz: mz reference value
        :type mz: float
        :param offset: offset values
        :type offset: float
        :return: dictionary of offset values
        :rtype: dict
        """
        self.offsets["offset_" + str(mz)] = offset

    def add_metadata(self, key, value):
        """
        Function to add the metadata of the spectrum  (NOT TESTED)

        :param key: attribute of the dictionary
        :type key: string
        :param value: value of the attribute
        :type value: string
        :return: dictionary of the metadata
        :rtype: dict
        """
        self.spectrummetadata[key] = value

    def get_intensity_at_mz(self, mz):
        """
        Function to obtain the intensity at a given mz value in the spectrum

        :param mz: centroid of interest
        :type mz: float
        :return: intensity at the given mz value
        :rtype: float
        """
        return self.intensities[np.where(self.mzs == mz)]

    def get_intensity_between_mzs(self, mzlow, mzhigh):
        """
        Function to calculate the intensity between two mz values. The intensity is the sum
        of all the intensities between these two values

        :param mzlow: left value of the interval
        :type mzlow: float
        :param mzhigh: right value of the interval
        :type mzhigh: float
        :return: Total intensity between two mz values
        :rtype: float
        """
        return np.sum(self.intensities[np.where((self.mzs >= mzlow) & (self.mzs <= mzhigh))])

    def get_offset(self, mz):
        """
        Function to calculate the offset of one mz value given a mz reference value

        :param mz: reference centroid of interest
        :type mz: float
        :return: offset of the mz value in the spectrum
        :rtype: dict
        """
        try:
            return self.offsets["offset_" + str(mz)]
        except:
            return np.nan



"""Class to select the peaks from the mean spectrum (peak picking routine)."""

"""___Built-In Modules___"""
from nectar_msi.gaussian_fitting.gaussian_fitting import GaussianFitting
from nectar_msi.plotting.plotting import Plotting
from nectar_msi.noise_correction.noise_determination import NoiseDetermination

"""___Third-Party Modules___"""
import numpy as np
import pandas as pd

"""___Authorship___"""
__author__ = "Ariadna Gonzalez-Fernandez"
__email__ = "ariadna.gonzalez@npl.co.uk"


class PeakPicking:
    def __init__(self):
        self.noisedetermination = NoiseDetermination()
        self.gaussian_fitting = GaussianFitting()
        self.plotting = Plotting()
        pass

    def peak_picking(self, mean_spectrum, path_outputs, plot_peaks=False, save_tables=False, save_fitting=False):
        """
        Function to select peaks in the mean spectrum. 1) selects all peaks by first derivative, 2) calculates resolving
         power to determine theoretical peak-width, 3) does gaussian fitting with constrained parameters, 4) creates
         final tables.
        :param mean_spectrum: mean spectrum used to find the peaks
        :type mean_spectrum: DataSpectrum
        :param path_outputs: path where to save the outputs
        :type path_outputs: str
        :param plot_peaks: option to plot the selected peaks
        :type plot_peaks: bool
        :param save_tables: option to save the tables as .csv files
        :type save_tables: bool
        :param save_fitting: option to save the gaussian fitting plots for all peaks
        :type save_fitting: bool
        :return: list of 200 most intense peaks, list of all selected peaks
        :rtype: pandas dataframe, pandas dataframe
        """

        print("Finding potential peaks by first derivatives...")
        # We select potential peaks by finding the first derivative (noise has been already masked):
        mz_mean = []
        peak_intensity_mean = []
        ppm = 30.0  # allowed distance between peaks. Farther than 30ppm are considered independent peaks.

        for i in range(2, len(mean_spectrum.intensities) - 2):
            # The noise is filtered out of the peak picking routine.
            if mean_spectrum.mzmask[i] == 0.0 and mean_spectrum.mzmask[i + 1] == 1.0:
                istart = i
            if mean_spectrum.mzmask[i] == 1.0 and mean_spectrum.mzmask[i + 1] == 0.0:
                spectrum_window = mean_spectrum.intensities[istart: i + 2]
                xaxis_window = mean_spectrum.mzs[istart: i + 2]
                # One extra value to each side of the window are added to avoid errors (Intensity = 0.).
                xaxis_window = np.append(2 * xaxis_window[0] - xaxis_window[1], xaxis_window)
                xaxis_window = np.append(xaxis_window, 2 * xaxis_window[len(xaxis_window) - 1]
                                         - xaxis_window[len(xaxis_window) - 2],)
                spectrum_window = np.append([0], spectrum_window)
                spectrum_window = np.append(spectrum_window, [0])

                # First derivatives are calculated in the window. These peaks are selected as potential peaks.
                dydx = np.diff(spectrum_window) / np.diff(xaxis_window)

                peaks_centroids = []
                peak_intensities = []
                for jjj in range(len(dydx) - 1):
                    if np.sign(dydx[jjj + 1]) < np.sign(dydx[jjj]):
                        local_mz = xaxis_window[jjj + 1]
                        local_intensity = spectrum_window[jjj + 1]
                        peaks_centroids = np.append(peaks_centroids, local_mz)
                        peak_intensities = np.append(peak_intensities, local_intensity)

                # If there is only one peak in the window, it is selected as potential peak.
                if len(peaks_centroids) == 1:
                    mz_mean = np.append(mz_mean, local_mz)
                    peak_intensity_mean = np.append(peak_intensity_mean, local_intensity)

                # If there are more than one peak in the window, the peaks are classified as independent if they are
                # located farther than 30ppm. The most intense peak is selected as reference if there is confusion.
                else:
                    for iii in range(len(peaks_centroids)):
                        intensities_near = peak_intensities[
                            np.where(np.abs(peaks_centroids - peaks_centroids[iii]) <
                                     (ppm * peaks_centroids[iii]) / 10**6)]

                        if max(intensities_near) == peak_intensities[iii]:
                            mz_mean = np.append(mz_mean, peaks_centroids[iii])
                            peak_intensity_mean = np.append(peak_intensity_mean, peak_intensities[iii])

        # To plot the selected peaks:
        if plot_peaks:
            self.plotting.plot_peak_picking(mean_spectrum, mz_mean, peak_intensity_mean)

        print("Number of detected peaks", len(mz_mean))
        print("Calculating resolving power...")
        # The 200 most intense peaks are chosen to estimate the resolving power of the dataset:
        peak_intensity_mean = np.argpartition(-peak_intensity_mean, 200)
        result_args = peak_intensity_mean[:200]  # This gives the index
        most_intense_peaks_mz = []
        for i in result_args:
            most_intense_peaks_mz = np.append(most_intense_peaks_mz, (mz_mean[i]))
        most_intense_peaks_mz = np.sort(most_intense_peaks_mz)
        # Gaussian fitting in the 200 most intense peaks is applied.
        # The correlation is calculated forcing the intercept to be at zero.
        list_of_fittings_subset = self.gaussian_fitting.gaussian_fitting_spectrum(
            mean_spectrum, most_intense_peaks_mz, path_outputs, save_fitting=save_fitting)
        # list_of_fittings is formed by:
        # 0          1               2       3               4           5           6
        # centroid   peak_intensity  width   chi2/Nsamples   Nsamples    Peak_type   Fitting_success

        mz_best_success = []
        width_best_success = []
        for succ in range(len(list_of_fittings_subset[:, 0])):
            if list_of_fittings_subset[:, 6][succ] == "True":
                mz_best_success = np.append(mz_best_success, float(list_of_fittings_subset[:, 0][succ]))
                width_best_success = np.append(width_best_success, float(list_of_fittings_subset[:, 2][succ]))

        # Correlation that gives the resolving power for the 200 most intense peaks. The function uses a sigma clipping
        # approximation to select only those peaks that are at 3sigma from the new-std with a tolerance of 0.01 from the
        # old-std. The intercept of the correlation is forced to be at zero.
        (func_best, distance_std_best, std_new_best, mz_corr_best, width_corr_best) = \
            self.noisedetermination.sigmaClip_stdv_zero_intercept(width_best_success, mz_best_success)

        print("Correlation most intense peaks:", func_best, std_new_best)
        self.plotting.plot_correlation(mz_corr_best, width_corr_best, func_best, mean_spectrum.mzs, path_outputs)

        # Once the resolving power is calculated, the boundaries of the width are constrained to +-3sigma the
        # theoretical value.
        # ref_m/z_mean centroid_gaussian peak_intensity width chi2/Nsamples Nsamples Peak_type Fitting_success"
        # 0            1                 2              3     4             5        6         7
        print("Creating final list of present peaks...")
        full_list_of_fittings = (self.gaussian_fitting.gaussian_fitting_spectrum_constrained(
            mean_spectrum, mz_mean, func_best, std_new_best, path_outputs, save_fitting=save_fitting))

        a = np.polyfit(full_list_of_fittings[:, 1].astype(float), full_list_of_fittings[:, 5].astype(float), 1)
        func_all_peaks = np.poly1d(a)
        # plots the resolving power for all peaks (upper limits applied on peak-width)
        self.plotting.plot_correlation_all_peaks(full_list_of_fittings[:, 1].astype(float),
                                                 full_list_of_fittings[:, 5].astype(float),
                                                 func_best, func_all_peaks, mean_spectrum.mzs, path_outputs)

        # Creates dataframe for 200 most intense peaks
        list_of_fittings_subset = pd.DataFrame(list_of_fittings_subset,
                                               columns=["gaussian centroid", "peak intensity", "width",
                                                        "chi2/Nsamples", "Nsamples", "Peak type", "Fitting success"])
        # Creates dataframe for all peaks
        full_list_of_fittings = pd.DataFrame(full_list_of_fittings,
                                             columns=["left min", "meas mz", "right min", "gaussian centroid",
                                                      "peak intensity", "width", "chi2/Nsamples", "Nsamples",
                                                      "Peak type", "S/N", "Fitting success", "Peak shape"])

        if save_tables:
            list_of_fittings_subset.to_csv(path_outputs + "list_of_fittings_subset.csv", index=False)
            full_list_of_fittings.to_csv(path_outputs + "full_list_of_fittings.csv", index=False)
            print("!List of peaks saved in: " + path_outputs)

        return list_of_fittings_subset, full_list_of_fittings
